@@ -11,6 +11,8 @@ import socket
 
 import attr
 
+import six
+
 from cryptoparser.common.algorithm import Authentication, KeyExchange
 from cryptoparser.common.exception import NotEnoughData, InvalidType, InvalidValue
 
@@ -118,6 +120,15 @@ class TlsHandshakeClientHelloAuthenticationECDSA(TlsHandshakeClientHelloAuthenti
         ])
 
 
+class TlsHandshakeClientHelloAuthenticationGOST(TlsHandshakeClientHelloAuthenticationBase):
+    def __init__(self, protocol_version, hostname):
+        super(TlsHandshakeClientHelloAuthenticationGOST, self).__init__(
+            protocol_version=protocol_version,
+            hostname=hostname,
+            authentication=Authentication.GOST_R3410_94
+        )
+
+
 class TlsHandshakeClientHelloAuthenticationRarelyUsed(TlsHandshakeClientHello):
     def __init__(self, protocol_version, hostname):
         _cipher_suites = TlsCipherSuiteVector([
@@ -338,8 +349,8 @@ class ClientPOP3(L7ClientStartTlsBase):
             response = self._l7_client._shortcmd('STLS')  # pylint: disable=protected-access
             if len(response) < 3 or response[:3] != b'+OK':
                 raise SecurityError(SecurityErrorType.UNSUPPORTED_SECURITY)
-        except poplib.error_proto:
-            raise SecurityError(SecurityErrorType.UNSUPPORTED_SECURITY)
+        except poplib.error_proto as e:
+            six.raise_from(SecurityError(SecurityErrorType.UNSUPPORTED_SECURITY), e)
 
     def _deinit_l7(self):
         try:
@@ -379,8 +390,8 @@ class ClientSMTP(L7ClientStartTlsBase):
             response, _ = self._l7_client.docmd('STARTTLS')
             if response != 220:
                 raise SecurityError(SecurityErrorType.UNSUPPORTED_SECURITY)
-        except smtplib.SMTPException:
-            raise SecurityError(SecurityErrorType.UNSUPPORTED_SECURITY)
+        except smtplib.SMTPException as e:
+            six.raise_from(SecurityError(SecurityErrorType.UNSUPPORTED_SECURITY), e)
 
     def _deinit_l7(self):
         try:
@@ -434,8 +445,8 @@ class ClientIMAP(L7ClientStartTlsBase):
             response, _ = self._l7_client.xatom('STARTTLS')
             if response != 'OK':
                 raise SecurityError(SecurityErrorType.UNSUPPORTED_SECURITY)
-        except imaplib.IMAP4.error:
-            raise SecurityError(SecurityErrorType.UNSUPPORTED_SECURITY)
+        except imaplib.IMAP4.error as e:
+            six.raise_from(SecurityError(SecurityErrorType.UNSUPPORTED_SECURITY), e)
 
     def _deinit_l7(self):
         try:
@@ -474,8 +485,8 @@ class ClientFTP(L7ClientStartTlsBase):
             response = self._l7_client.sendcmd('AUTH TLS')
             if not response.startswith('234'):
                 raise SecurityError(SecurityErrorType.UNSUPPORTED_SECURITY)
-        except ftplib.all_errors:
-            raise SecurityError(SecurityErrorType.UNSUPPORTED_SECURITY)
+        except ftplib.all_errors as e:
+            six.raise_from(SecurityError(SecurityErrorType.UNSUPPORTED_SECURITY), e)
 
     def _deinit_l7(self):
         try:
@@ -511,10 +522,10 @@ class ClientRDP(L7ClientStartTlsBase):
             cotp = COTPConnectionConfirm.parse_exact_size(tpkt.message)
             neg_rsp = RDPNegotiationResponse.parse_exact_size(cotp.user_data)
             self.l4_transfer.flush_buffer(len(request_bytes))
-        except socket.timeout:
-            raise SecurityError(SecurityErrorType.UNSUPPORTED_SECURITY)
-        except (InvalidValue, InvalidType):
-            raise SecurityError(SecurityErrorType.UNSUPPORTED_SECURITY)
+        except socket.timeout as e:
+            six.raise_from(SecurityError(SecurityErrorType.UNSUPPORTED_SECURITY), e)
+        except (InvalidValue, InvalidType) as e:
+            six.raise_from(SecurityError(SecurityErrorType.UNSUPPORTED_SECURITY), e)
 
         if RDPProtocol.SSL not in neg_rsp.protocol:
             raise SecurityError(SecurityErrorType.UNSUPPORTED_SECURITY)
@@ -523,6 +534,66 @@ class ClientRDP(L7ClientStartTlsBase):
         pass
 
 
+@attr.s
+class ClientXMPP(L7ClientStartTlsBase):
+    _STREAM_OPEN = (
+        '<stream:stream xmlns=\'jabber:client\' xmlns:stream=\'http://etherx.jabber.org/streams\' '
+        'xmlns:tls=\'http://www.ietf.org/rfc/rfc2595.txt\' to=\'{}\' xml:lang=\'en\' version=\'1.0\'>'
+    )
+    _STARTTLS = b'<starttls xmlns=\'urn:ietf:params:xml:ns:xmpp-tls\'/>'
+
+    @classmethod
+    def get_scheme(cls):
+        return 'xmpp'
+
+    @classmethod
+    def get_default_port(cls):
+        return 5222
+
+    @staticmethod
+    def _init_xmpp(l4_transfer, address):
+        stream_open_message = ClientXMPP._STREAM_OPEN.format(address).encode("utf-8")
+        l4_transfer.send(stream_open_message)
+
+        l4_transfer.receive_until(b'<stream:')
+        l4_transfer.receive_until(b'>')
+
+        if b'stream:error' in l4_transfer.buffer:
+            raise NetworkError(NetworkErrorType.NO_RESPONSE)
+
+        if b'stream:features' not in l4_transfer.buffer:
+            l4_transfer.receive_until(b'</stream:features>')
+
+        if b'<starttls xmlns=\'urn:ietf:params:xml:ns:xmpp-tls\'>' not in l4_transfer.buffer:
+            raise SecurityError(SecurityErrorType.UNSUPPORTED_SECURITY)
+
+        l4_transfer.flush_buffer()
+
+        l4_transfer.send(ClientXMPP._STARTTLS)
+        l4_transfer.receive_until(b'>')
+
+        if b'stream:error' in l4_transfer.buffer:
+            raise SecurityError(SecurityErrorType.UNSUPPORTED_SECURITY)
+
+        if l4_transfer.buffer != b'<proceed xmlns=\'urn:ietf:params:xml:ns:xmpp-tls\'/>':
+            raise SecurityError(SecurityErrorType.UNSUPPORTED_SECURITY)
+
+        l4_transfer.flush_buffer()
+
+    def _init_l7(self):
+        self._l7_client = L7ClientTls(self.address, self.port, self.timeout)
+        self._l7_client.init_connection()
+        self.l4_transfer = self._l7_client.l4_transfer
+
+        try:
+            self._init_xmpp(self.l4_transfer, self.address)
+        except NotEnoughData as e:
+            six.raise_from(NetworkError(NetworkErrorType.NO_RESPONSE), e)
+
+    def _deinit_l7(self):
+        pass
+
+
 class TlsClient(object):
     _last_processed_message_type = attr.ib(init=False, default=None)
     server_messages = attr.ib(init=False, default={})
@@ -595,11 +666,11 @@ class TlsClientHandshake(TlsClient):
 
             try:
                 transfer.receive(receivable_byte_num)
-            except NotEnoughData:
+            except NotEnoughData as e:
                 if transfer.buffer:
-                    raise NetworkError(NetworkErrorType.NO_CONNECTION)
+                    six.raise_from(NetworkError(NetworkErrorType.NO_CONNECTION), e)
 
-                raise NetworkError(NetworkErrorType.NO_RESPONSE)
+                six.raise_from(NetworkError(NetworkErrorType.NO_RESPONSE), e)
 
 
 @attr.s
@@ -646,7 +717,7 @@ class SslClientHandshake(TlsClient):
 
             try:
                 transfer.receive(receivable_byte_num)
-            except NotEnoughData:
+            except NotEnoughData as e:
                 if transfer.buffer:
                     try:
                         tls_record = TlsRecord.parse_exact_size(transfer.buffer)
@@ -660,8 +731,8 @@ class SslClientHandshake(TlsClient):
                                     TlsAlertDescription.HANDSHAKE_FAILURE,
                                     TlsAlertDescription.INTERNAL_ERROR,
                                 ])):
-                            raise NetworkError(NetworkErrorType.NO_RESPONSE)
+                            six.raise_from(NetworkError(NetworkErrorType.NO_RESPONSE), e)
 
-                        raise NetworkError(NetworkErrorType.NO_CONNECTION)
+                        six.raise_from(NetworkError(NetworkErrorType.NO_CONNECTION), e)
                 else:
-                    raise NetworkError(NetworkErrorType.NO_RESPONSE)
+                    six.raise_from(NetworkError(NetworkErrorType.NO_RESPONSE), e)
